import numpy as np
import sys
from  ..population.binary_generator import  BNS_gaussian_distribution,BNS_uniform_distribution,BBH_mass_distribution
from scipy.interpolate import interp1d, interp2d
from scipy.stats import ncx2, gaussian_kde
from scipy.integrate import quad
from tqdm import tqdm
import pandas as pd
from ..statistic.likelihoods import log_CF_likelihood

def PdetGW_PdetEM_compute_area(EM_Pdet, detectors, sky_area,
SNR_net_thr=12., Nsample=5000, mass_generation='BNS-gaussian', alpha=1.6):

    #################################
    print('I am computing the Pdet_EM detection probability \n')
    df=pd.read_pickle(EM_Pdet)
    dL = df['dL'][:4000]
    cosi = df['cosi'][:4000]
    d_array= np.linspace(np.min(dL),np.max(dL),400)
    cosiota_array= np.linspace(-1*np.max(cosi),np.max(cosi),400)
    dd, cc = np.meshgrid(d_array,cosiota_array)
    pos = np.stack([dd.flatten(),cc.flatten()])
    dL=np.append(dL,
         [dL, dL, dL,
          -dL, -dL, -dL, -dL])
    cosi=np.append(cosi,
           [-1*cosi,cosi-2.,-cosi+2,
            cosi,-cosi,cosi-2.,-cosi+2.])
    to_fit = np.vstack([dL,cosi])
    agp_det = gaussian_kde(to_fit)
    values = np.reshape(agp_det(pos).T, dd.shape)
    norm_factor=np.max(values)
    values = values/np.max(values) # Put pdet to 1 when you detect with Prob 1
    PdetEM=interp2d(d_array,cosiota_array, values, kind='cubic', bounds_error=False, fill_value=0.)
    #Above computes the PdetEM in dL (Mpc) and cosiota
    #################################

    if mass_generation == 'BNS-gaussian':
            m1, m2 = BNS_gaussian_distribution(Nsample, mean=1.35, sigma=0.15)
    if mass_generation == 'BNS-uniform':
            m1, m2 = BNS_uniform_distribution(Nsample, mmin=1.2, mmax=1.6)
    if mass_generation == 'BBH-powerlaw':
            m1, m2 = BBH_mass_distribution(Nsample, mmin=5., mmax=50., alpha=self.alpha)

    RAs = np.random.rand(Nsample)*2.0*np.pi # R.A. in radians
    r = np.random.rand(Nsample)
    Decs = np.arcsin(2.0*r - 1.0) # Declination in radians
    q = np.random.rand(Nsample)
    incs = np.arccos(2.0*q - 1.0) # Source inclination in radians
    psis = np.random.rand(Nsample)*2.0*np.pi # Psi in radians
    det_prob=[]
    #Fix a tgps to the one og GW170817
    tgps=1187008882.43

    lum_val=.1 # Start step
    interpolated_integrand=None

    print('Starting to compute the Pdet conditioned \n')
    while True: #Thr here decided according to the precision of the monte carlo

        SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis,incs,lum_val,interpolated_integrand=interpolated_integrand)
        SNR_detected = np.sqrt(ncx2.rvs(2*len(detectors), SNR_net**2.))
        localization_area = 10.0**(5.06)/SNR_detected**2 # 90% localization of the event in deg^2

        # Detected GW with good sky-loc
        NdetBin= np.where((localization_area<=sky_area) &
        (SNR_detected>=SNR_net_thr))[0]

        cosidet_GW=np.cos(incs[NdetBin])
        dLdet_GW=np.ones_like(cosidet_GW)*lum_val
        gpos=np.vstack([dLdet_GW,cosidet_GW])
        Pdet_EM_conditioned=agp_det(gpos)/norm_factor

        det_prob.append(np.sum(Pdet_EM_conditioned)/Nsample)
        dcrit = lum_val
        if det_prob[-1]<(1./Nsample):
            break
        lum_val+=2 # The Mpc step

        print('Loop stopping :{:e}>{:e}'.format(det_prob[-1],1./Nsample),end='\r')

    det_prob = np.array(det_prob)
    d_lum = np.linspace(.1,lum_val,len(det_prob))
    PdetGW_EM_cond = interp1d(d_lum, det_prob,bounds_error=False,fill_value=0.)

    return PdetGW_EM_cond, dcrit

def detected_binaries_EM(binaries_dict,distance_generator,sigma_d,sigma_iota,EM_Pdet=None,true_value_mode=True):

    df=pd.read_pickle(EM_Pdet)
    dL = df['dL'][:4000]
    cosi = df['cosi'][:4000]
    d_array= np.linspace(np.min(dL),np.max(dL),400)
    cosiota_array= np.linspace(-1*np.max(cosi),np.max(cosi),400)
    dd, cc = np.meshgrid(d_array,cosiota_array)
    pos = np.stack([dd.flatten(),cc.flatten()])
    dL=np.append(dL,
         [dL, dL, dL,
          -dL, -dL, -dL, -dL])
    cosi=np.append(cosi,
           [-1*cosi,cosi-2.,-cosi+2,
            cosi,-cosi,cosi-2.,-cosi+2.])
    to_fit = np.vstack([dL,cosi])
    agp_det = gaussian_kde(to_fit)
    values = np.reshape(agp_det(pos).T, dd.shape)
    values = values/np.max(values) # Put pdet to 1 when you detect with Prob 1
    PdetEM=interp2d(d_array,cosiota_array, values, kind='cubic', bounds_error=False, fill_value=0.)

    tgps = binaries_dict['tgps']
    m1 = binaries_dict['mass1']
    m2 = binaries_dict['mass2']
    RAs = binaries_dict['RA']
    Decs = binaries_dict['dec']
    psis = binaries_dict['psi']
    incs = binaries_dict['iota']
    lum_val=[]

    for zz in binaries_dict['z']:
        lum_val.append(distance_generator(zz))
    lum_val = np.array(lum_val)

    det_iota_EM=np.zeros_like(lum_val)
    det_distance_EM=np.zeros_like(lum_val)
    EM_detected = []

    for i_sig in range(len(det_distance_EM)):

        det_iota_EM[i_sig] = np.random.randn()*sigma_iota+incs[i_sig]
        det_distance_EM[i_sig] = np.random.randn()*sigma_d+lum_val[i_sig]

        while det_distance_EM[i_sig]<=0:
            det_distance_EM[i_sig] = np.random.randn()*sigma_d+lum_val[i_sig]

        if true_value_mode:
            det_trial=PdetEM(lum_val[i_sig],np.cos(incs[i_sig]))
        else:
            det_trial=PdetEM(det_distance_EM[i_sig],np.cos(det_iota_EM[i_sig]))

        EM_detected.append(np.random.rand()<det_trial[0])
    EM_detected = np.array(EM_detected)

    binaries_dict['EM_detected'] = EM_detected
    binaries_dict['det_iota_EM'] = det_iota_EM
    binaries_dict['det_luminosity_distance_EM'] = det_distance_EM

    return binaries_dict

def detected_binaries_GW(binaries_dict,distance_generator,detectors,mode='standard',SNR_thr=12.):
    """ This function generates N binaries and return them as a dict of detections
    Parameters
    ----------
    binaries_dict: dictionary
        Dictionary containing the binaries parameters.  RA, dec,iota,psi,mass1,mass2,tgps,z
    distance_generator: lambda function
        lambda function in z for obtaining luminosity distance in Mpc
    detectors: list
        list of detector objects
    SNR_thr: float
        threshold for GW SNR
    EM_Pdet: str(optional)
        String with the pickle file containing the EM Pdet

    Returns
    -------
    binaries: dictionary
        Same input dict but with extra fields. It will also tell you if a binary is observed
    """
    tgps = binaries_dict['tgps']
    m1 = binaries_dict['mass1']
    m2 = binaries_dict['mass2']
    RAs = binaries_dict['RA']
    Decs = binaries_dict['dec']
    psis = binaries_dict['psi']
    incs = binaries_dict['iota']
    lum_val=[]
    for zz in binaries_dict['z']:
        lum_val.append(distance_generator(zz))
    lum_val = np.array(lum_val)

    SNR_net, interpolated_integrand = matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis,incs,lum_val)

    if mode == 'd-only':
        print('Using mode: '+mode)
        print('Please remember that with this method. You have to use L(x|d)=N(d,1.8/rho)\n')
        detected_distance=(1.8/SNR_net)*np.random.randn(len(SNR_net))+lum_val
        detected_iota=incs
        SNR_detected=(lum_val/detected_distance)*SNR_net
    elif mode == 'no-noise':
        print('Please remember that with this method, your GW likelihood is a delta.\n')
        SNR_detected=SNR_net
        detected_distance=lum_val
        detected_iota=incs

    elif mode == 'CF':
        CF = log_CF_likelihood(detectors)

        lums_det=np.ones_like(SNR_net)
        iota_det=np.ones_like(SNR_net)
        print('Simulating detected binaries \n')
        for i_sig in tqdm(range(len(SNR_net))):

            signal={'tgps': np.array([tgps[i_sig]]),
            'dec': np.array([Decs[i_sig]]),
            'RA': np.array([RAs[i_sig]]),
            'psi': np.array([psis[i_sig]]),
            'iota': np.array([incs[i_sig]]),
            'SNR_detected': np.array([SNR_net[i_sig]])}

            new_sample = CF.sample(0.2,2,signal,Nsample=1, bound=True)
            lums_det[i_sig]=lum_val[i_sig]*new_sample[0][0]
            iota_det[i_sig]=np.arccos(new_sample[1][0])

        SNR_detected, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,
        psis,iota_det,lums_det,interpolated_integrand=interpolated_integrand)
        detected_distance=lums_det
        detected_iota=iota_det
    elif mode == 'standard':
        print('Using mode: '+mode)
        SNR_detected = np.sqrt(ncx2.rvs(df=2*len(detectors), nc=SNR_net**2.))

    GW_detected = SNR_detected>SNR_thr

    binaries_dict['det_iota'] = detected_iota
    binaries_dict['det_luminosity_distance'] = detected_distance
    binaries_dict['luminosity_distance'] = lum_val
    binaries_dict['SNR_optimal'] = SNR_net
    binaries_dict['SNR_detected'] = SNR_detected
    binaries_dict['GW_detected'] = GW_detected

    return binaries_dict

def matched_filter_SNR(tgps, m1,m2,detectors,RAs,Decs,psis,incs,lum_val,interpolated_integrand=None):
    """A function that computes the mateched filter SNR_net

    Parameters
    ----------
    tgps: numpy array
        Arrival time of the signal at the SSB
    m1: numpy array
        Mass of the 1st object
    m2: numpy array
        Mass of the 2nd object
    detectors: list
        List of detector objects
    RAs: numpy array
        Right ascension in radians of the detection
    Decs: numpy array
        Declination in radians of the detections
    incs: numpy array
        Iota of the detections in radiant
    lum_val: numpy array
        Luminosity distance in MPC
    interpolated_integrand: list (optiona) of interpolation
        A list containint the integral of f^{-7/3} up a given freq evaluated as an interpolation

    Returns
    -------
    SNR_network_square: np array
        SNR of the event at the detectors
    interpolated_integrand: list (optiona) of interpolation
        A list containint the integral of f^{-7/3} up a given freq evaluated as an interpolation
    """

    G_cost=6.674e-11 # G in m3 kg-1 s-2
    c_cost= 299792458.0 # Speed of light in m/s
    conversion_MPC_m=3.086e+22 #Conversion Mpc m
    conversion_solarmass_kg=1.988e30 # Conversion from solar masses in kgs
    chirp_mass_source_frame = ((m1*m2)**(3./5))/(m1+m2)**(1./5) # Chirp mass
    mergers_fISCO=2.2e3/(chirp_mass_source_frame*2.0**(6./5.))
    freq_det=np.logspace(np.log10(10),np.log10(2048),500)

    if interpolated_integrand is None:
        psd_det = [detectors[i_det].compute_PSD() for i_det in range(len(detectors))]
        function_to_integrate=[lambda freq_value: freq_value**(-7./3)/(psd_det[i_det](freq_value))**2.
                               for i_det in range(len(detectors))]
        freq_integration=np.logspace(np.log10(20),np.log10(2048),50)

        integrated_function=[np.array([quad(function_to_integrate[i_det],20,i_freq)[0]
                                       for i_freq in freq_integration]) for i_det in range(len(detectors))]
        interpolated_integrand=[interp1d(freq_integration,integrated_function[i_det])
                                for i_det in range(len(detectors))]

    Fplus_Fcross=[detectors[i_det].compute_antenna_patterns(tgps, RAs, Decs, psis)
                         for i_det in range(len(detectors))]
    overall_amplitude=(np.sqrt(5./(24*np.pi**(4./3.)))*np.power(G_cost*chirp_mass_source_frame*conversion_solarmass_kg,5./6)
                              /np.power(c_cost,3./2))
    overall_amplitude=overall_amplitude/(lum_val*conversion_MPC_m)

    reduced_plus_amplitude=overall_amplitude*((1+np.cos(incs)**2.)/2.)
    reduced_cross_amplitude=overall_amplitude*np.cos(incs)

    SNRsquare=[4*((reduced_plus_amplitude*Fplus_Fcross[i_det][0])**2+
                  (reduced_cross_amplitude*Fplus_Fcross[i_det][1])**2)*interpolated_integrand[i_det](mergers_fISCO)
               for i_det in range(len(detectors))]

    SNR_network_square=np.sum(np.array(SNRsquare),axis=0)

    return np.sqrt(SNR_network_square),interpolated_integrand

def PdetGW_compute(fast_method,horizon = None, detectors = None, SNR_net_thr=12.,
mass_generation='BNS-gaussian', alpha=1.6, Nsample=5000,mode='standard'):
    """Returns the detection probability as a function of the luminosity distance Lambda function

    Parameters
    ----------
    fast_method : boolean
        True if you want Pdet as a box
    horizon : float
        Horizon for the fast_method
    detectors : List
        List of detector objects
    SNR_net_thr : float
        SNR threshold for detection
    mass_generation : str
        Masses generator to use "BNS_gaussian_distribution", "BNS_uniform_distribution"
        "BBH_mass_distribution"
    alpha : float
        Slope for the power low mass distribution of BBHs
    Nsample : integer
        Number of binaries to simulate

    Returns
    -------
    Pdet: lambda function
            Detection probability
    dcrit: float
        Horizon or distance at which the simulation stops. This is given when you stop detecting binaries.
    """
    if fast_method:
        d_lum=np.linspace(0.,horizon,1000,endpoint=True)
        det_prob=np.ones_like(d_lum)
        PdetGW = interp1d(d_lum, det_prob,bounds_error=False,fill_value=0.)
        dcrit = horizon
    else:
        if mass_generation == 'BNS-gaussian':
                m1, m2 = BNS_gaussian_distribution(Nsample, mean=1.35, sigma=0.15)
        if mass_generation == 'BNS-uniform':
                m1, m2 = BNS_uniform_distribution(Nsample, mmin=1.2, mmax=1.6)
        if mass_generation == 'BBH-powerlaw':
                m1, m2 = BBH_mass_distribution(Nsample, mmin=5., mmax=50., alpha=self.alpha)


        RAs = np.random.rand(Nsample)*2.0*np.pi # R.A. in radians
        r = np.random.rand(Nsample)
        Decs = np.arcsin(2.0*r - 1.0) # Declination in radians
        q = np.random.rand(Nsample)
        incs = np.arccos(2.0*q - 1.0) # Source inclination in radians
        psis = np.random.rand(Nsample)*2.0*np.pi # Psi in radians
        det_prob=[]
        #Fix a tgps to the one og GW170817
        tgps=1187008882.43

        lum_val=.1 # Step in Mpc
        interpolated_integrand=None
        print('Using mode: '+mode)
        print('Be aware that your GW likelihoods must use the same thing')

        while True: #Thr here decided according to the precision of the monte carlo
            SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis,incs,lum_val,interpolated_integrand=interpolated_integrand)

            if mode == 'd-only':
                detected_distance=(1.8/SNR_net)*np.random.randn(len(SNR_net))+lum_val
                SNR_detected=(lum_val/detected_distance)*SNR_net
                det_prob.append(len(np.where(SNR_detected>SNR_net_thr)[0])/Nsample)
            elif mode == 'no-noise':
                SNR_detected=SNR_net
                det_prob.append(len(np.where(SNR_detected>SNR_net_thr)[0])/Nsample)
            elif mode == 'CF':
                survival = ncx2.sf(SNR_net_thr**2., 2*len(detectors), SNR_net**2.)
                det_prob.append(np.sum(survival)/Nsample)
            elif mode == 'standard':
                survival = ncx2.sf(SNR_net_thr**2.,  2*len(detectors), SNR_net**2.)
                det_prob.append(np.sum(survival)/Nsample)

            dcrit = lum_val
            if det_prob[-1]<(1./Nsample):
                break
            lum_val+=.1

        det_prob = np.array(det_prob)
        d_lum = np.linspace(.1,lum_val,len(det_prob))
        PdetGW = interp1d(d_lum, det_prob,bounds_error=False,fill_value=0.)

    return PdetGW, dcrit


def PdetGW_compute_cosiota(detectors = None, SNR_net_thr=12.,
mass_generation='BNS-gaussian', alpha=1.6, Nsample=5000):
    """Returns the GW detection probability on the D cosiota plane

    Parameters
    ----------
    detectors : List
        List of detector objects
    SNR_net_thr : float
        SNR threshold for detection
    mass_generation : str
        Masses generator to use "BNS_gaussian_distribution", "BNS_uniform_distribution"
        "BBH_mass_distribution"
    alpha : float
        Slope for the power low mass distribution of BBHs
    Nsample : integer
        Number of binaries to simulate

    Returns
    -------
    Pdet: lambda function
            Detection probability in dl - iota
    dcrit: float
        Horizon or distance at which the simulation stops. This is given when you stop detecting binaries.
    """
    if mass_generation == 'BNS-gaussian':
            m1, m2 = BNS_gaussian_distribution(Nsample, mean=1.35, sigma=0.15)
    if mass_generation == 'BNS-uniform':
            m1, m2 = BNS_uniform_distribution(Nsample, mmin=1.2, mmax=1.6)
    if mass_generation == 'BBH-powerlaw':
            m1, m2 = BBH_mass_distribution(Nsample, mmin=5., mmax=50., alpha=self.alpha)

    RAs = np.random.rand(Nsample)*2.0*np.pi # R.A. in radians
    r = np.random.rand(Nsample)
    Decs = np.arcsin(2.0*r - 1.0) # Declination in radians
    q = np.random.rand(Nsample)
    psis = np.random.rand(Nsample)*2.0*np.pi # Psi in radians
    incs = np.arccos(2.0*q - 1.0) # Source inclination in radians
    det_prob=[]
    #Fix a tgps to the one og GW170817
    tgps=1187008882.43

    lum_val=.1 # Step in Mpc
    interpolated_integrand=None
    # The loop below is used to find an estimation of the GW horizon
    print('Computing the Luminosity distance horizon')
    while True: #Thr here decided according to the precision of the monte carlo
        SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis,incs,lum_val,interpolated_integrand=interpolated_integrand)
        survival = ncx2.sf(SNR_net_thr**2., 2*len(detectors), SNR_net**2.)

        det_prob.append(np.sum(survival)/Nsample)
        dcrit = lum_val
        if np.sum(survival)<1.:
            break
        lum_val+=.1

    incs_array=np.linspace(0,np.pi,100)
    lum_distance=np.linspace(0,dcrit,100)
    d_lum_mat, incs_mat = np.meshgrid(lum_distance,incs_array)

    detection_prob = np.zeros_like(incs_mat)
    print('Computing the cosiota-d detection probability')

    for indx, _ in tqdm(np.ndenumerate(detection_prob),total=len(detection_prob.flatten())):
        SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis
        ,incs_mat[indx],d_lum_mat[indx],interpolated_integrand=interpolated_integrand)
        survival = ncx2.sf(SNR_net_thr**2., 2*len(detectors), SNR_net**2.)
        detection_prob[indx] = np.sum(survival)/Nsample

    interpolant = interp2d(d_lum_mat[0,:], incs_mat[:,0], detection_prob, kind='cubic', bounds_error=False, fill_value=0.)

    return interpolant, dcrit



def PdetGW_compute_cosiota(detectors = None, SNR_net_thr=12.,
mass_generation='BNS-gaussian', alpha=1.6, Nsample=5000):
    """Returns the GW detection probability on the D cosiota plane. Does not multiply the prior for cosiotas.

    Parameters
    ----------
    detectors : List
        List of detector objects
    SNR_net_thr : float
        SNR threshold for detection
    mass_generation : str
        Masses generator to use "BNS_gaussian_distribution", "BNS_uniform_distribution"
        "BBH_mass_distribution"
    alpha : float
        Slope for the power low mass distribution of BBHs
    Nsample : integer
        Number of binaries to simulate

    Returns
    -------
    Pdet: interp2d
            Detection probability in terms of (d,iota)
    dcrit: float
        Horizon or distance at which the simulation stops. This is given when you stop detecting binaries.
    """
    if mass_generation == 'BNS-gaussian':
            m1, m2 = BNS_gaussian_distribution(Nsample, mean=1.35, sigma=0.15)
    if mass_generation == 'BNS-uniform':
            m1, m2 = BNS_uniform_distribution(Nsample, mmin=1.2, mmax=1.6)
    if mass_generation == 'BBH-powerlaw':
            m1, m2 = BBH_mass_distribution(Nsample, mmin=5., mmax=50., alpha=self.alpha)

    RAs = np.random.rand(Nsample)*2.0*np.pi # R.A. in radians
    r = np.random.rand(Nsample)
    Decs = np.arcsin(2.0*r - 1.0) # Declination in radians
    q = np.random.rand(Nsample)
    psis = np.random.rand(Nsample)*2.0*np.pi # Psi in radians
    incs = np.arccos(2.0*q - 1.0) # Source inclination in radians
    det_prob=[]
    #Fix a tgps to the one og GW170817
    tgps=1187008882.43

    lum_val=.1 # Step in Mpc
    interpolated_integrand=None
    # The loop below is used to find an estimation of the GW horizon
    print('Computing the Luminosity distance horizon')
    while True: #Thr here decided according to the precision of the monte carlo
        SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis,incs,lum_val,interpolated_integrand=interpolated_integrand)
        survival = ncx2.sf(SNR_net_thr**2., 2*len(detectors), SNR_net**2.)

        det_prob.append(np.sum(survival)/Nsample)
        dcrit = lum_val
        if np.sum(survival)<1.:
            break
        lum_val+=.1

    incs_array=np.linspace(0,np.pi,100)
    lum_distance=np.linspace(0,dcrit,100)
    d_lum_mat, incs_mat = np.meshgrid(lum_distance,incs_array)

    detection_prob = np.zeros_like(incs_mat)
    print('Computing the cosiota-d detection probability')

    for indx, _ in tqdm(np.ndenumerate(detection_prob),total=len(detection_prob.flatten())):
        SNR_net, interpolated_integrand=matched_filter_SNR(tgps,m1,m2,detectors,RAs,Decs,psis
        ,incs_mat[indx],d_lum_mat[indx],interpolated_integrand=interpolated_integrand)
        survival = ncx2.sf(SNR_net_thr**2., 2*len(detectors), SNR_net**2.)
        detection_prob[indx] = np.sum(survival)/Nsample

    Pdet = interp2d(d_lum_mat[0,:], incs_mat[:,0], detection_prob, kind='cubic', bounds_error=False, fill_value=0.)

    return Pdet, dcrit
