import numpy as np
import bilby
from tqdm import tqdm
from scipy.stats import gaussian_kde



def combine_posteriors_analytic(list_files,list_vars):
    """ This function combine the posteriors by multiplying them

    Parameters
    ----------
    list_files: list of strings
        List containing the bilby result files to read
    dict_vars: dictionary
        dict with fields set to the name of vars. Inside the fields a [min, max] array
        
    Returns
    -------
    Htot: matrix
        Matrix containing the posterior on the grid
    grid_list: list
        List containing the
    """

    Ninj=len(list_files)
        
    g_list=[]
    for ind in tqdm(range(Ninj)):
        file_to_read=list_files[ind]
        result = bilby.result.read_in_result(filename=file_to_read)
        mat_fit=np.vstack([result.posterior[var].values for var in list_vars])
        gfit=gaussian_kde(mat_fit)
        g_list.append(gfit)
        
    log_pdf=lambda variab:loop_on_g(g_list,variab)
              
    return g_list

def loop_on_g(g_list,variab):
    
    Ninj=len(g_list)
    logpdf=0.
    for indx in range(Ninj):
        logpdf+=g_list[indx].logpdf(variab)
    return logpdf


def combine_posteriors(list_files,dict_vars,bins):
    """ This function combine the posteriors by multiplying them

    Parameters
    ----------
    list_files: list of strings
        List containing the bilby result files to read
    dict_vars: dictionary
        dict with fields set to the name of vars. Inside the fields a [min, max] array
    bins: integer

    Returns
    -------
    Htot: matrix
        Matrix containing the posterior on the grid
    grid_list: list
        List containing the
    """

    list_vars=list(dict_vars.keys())
    Ninj=len(list_files)
    g_list=[]
    Htot=1

    list_edges=[]
    for var in list_vars:
        list_edges.append(np.linspace(dict_vars[var][0],dict_vars[var][1],bins,endpoint=True))

    list_central=[]
    for i in range(len(list_vars)):
        list_central.append(list_edges[i][0:-1]+0.5*(list_edges[i][1]-list_edges[i][0]))

    grid_list=np.meshgrid(*list_central,indexing='ij')
    grid_dict={}

    i=0
    for var in list_vars:
        grid_dict[var]=grid_list[i]
        i+=1


    for ind in tqdm(range(Ninj)):
        file_to_read=list_files[ind]
        result = bilby.result.read_in_result(filename=file_to_read)
        mat_fit=np.vstack([result.posterior[var].values for var in list_vars])

        gfit=gaussian_kde(mat_fit)
        resamp=gfit.resample(10000000)
        H,_=np.histogramdd(np.transpose(resamp),bins=list_edges)
        Htot*=H
        Htot/=Htot.sum()

    return Htot,grid_dict
